package Anarres::Mud::Driver::Program;

use strict;
use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS %PROGS);
use Exporter;
use Carp qw(:DEFAULT cluck);
use Data::Dumper;
use File::Basename;
use String::Escape qw(quote printable);
use Anarres::Mud::Driver::Program::Efun qw(%EFUNS);

@ISA = qw(Exporter);
{
	my @flags = qw(M_NOMASK M_NOSAVE
					M_PRIVATE M_PROTECTED M_PUBLIC
					M_VARARGS);
	@EXPORT_OK = (@flags, qw(package_to_path path_to_package));
	%EXPORT_TAGS = (
		flags	=> \@flags,
		q[:ALL]	=> \@EXPORT_OK,
			);
}

%PROGS = (
	"/foo/bar"	=> new Anarres::Mud::Driver::Program(Path=>"/foo/bar"),
		);

sub M_NOMASK	() { 0x001 }
sub M_NOSAVE	() { 0x002 }
sub M_PRIVATE	() { 0x004 }
sub M_PROTECTED	() { 0x008 }
sub M_PUBLIC	() { 0x010 }

sub M_VARARGS	() { 0x020 }

sub M_EFUN		() { 0x100 }
sub M_INHERITED	() { 0x200 }

# Class methods

sub new {
	my $class = shift;
	my $self = ($#_ == 0) ? { %{ (shift) } } : { @_ };

	confess "No Path in program" unless $self->{Path};

	$self->{Inherits} = { };
	$self->{Globals} = { };
	$self->{Methods} = { %EFUNS };
	$self->{MethodFlags} = { map { $_ => M_EFUN } keys %EFUNS };

	$self->{ScopeStack} = [ ];

	return bless $self, $class;
}

sub find {	# find Anarres::Mud::Driver::Program $path
	return $PROGS{$_[1]};
}

sub path_to_package {
	my $path = shift;
	$path =~ s,/,::,g;
	$path =~ s/\.c$//;
	return "Anarres::Mud::Library" . $path;
}

sub package_to_path {
	my $package = shift;
	die "package_to_path: Invalid package name"
			unless $package =~ s/^Anarres::Mud::Library//;
	$package =~ s,::,/,g;
	return $package;
}

# Instance query methods

sub path { return $_[0]->{Path}; }
sub source { return $_[0]->{Source}; }
sub ppsource { return $_[0]->{PPSource}; }
sub package { return path_to_package $_[0]->{Path}; }

sub globals { return values %{ $_[0]->{Globals} }; }
sub methods { return values %{ $_[0]->{Methods} }; }

sub variable {
	my ($self, $name) = @_;

	my $var;

	foreach (@{$self->{ScopeStack}}) {
		if ($_->{$name}) {
			return $_->{$name};
		}
	}

	if ($self->{Globals}->{$name}) {
		return $self->{Globals}->{$name};
	}

	return undef;
}

# Instance modification methods

sub scope_start {
	my ($self) = @_;
	unshift(@{$self->{ScopeStack}}, { } );
}

sub scope_end {
	my ($self) = @_;
	shift(@{$self->{ScopeStack}});
}

sub local {
	my ($self, $var) = @_;

	die "var '$var' not blessed" unless ref($var) =~ /::/;

	my $name = $var->name;

	return "Local $name masks previous definition"
			if $self->variable($name);

	# print "Storing local variable " . $var->dump . "\n";

	$self->{ScopeStack}->[0]->{$name} = $var;

	return ();
}

sub global {
	my ($self, $var, $mods) = @_;

	die "var '$var' not blessed" unless ref($var) =~ /::/;

	my $name = $var->name;

	return "Global $name masks previous definition in file XXX"
			if $self->{Globals}->{$name};

	# print "Storing variable $name\n";

	$self->{Globals}->{$name} = $var;

	return ();
}

sub method {
	my ($self, $method, $mods) = @_;

	unless (ref $method) {
		$method =~ s/^.*:://;	# XXX Remove and do properly.
		my $ob = $self->{Methods}->{$method};
		die "Method $method not found" unless $ob;
		return $ob;
	}

	my $name = $method->name;

	return "Method $name already exists"
			if $self->{Methods}->{$name};

	# XXX Check prototype match with superclass
	# XXX Check sanity of modifiers

	$self->{Methods}->{$name} = $method;
	$self->{MethodFlags}->{$name} = $mods;

	return ();
}

sub inherit {
	my ($self, $path) = @_;

	my $inh = $PROGS{$path};
	return "Could not find inherited program $_" unless $inh;

	my $name = basename($path, ".c");
	return "Already inheriting file named $name"
					if $self->{Inherits}->{$path};

	$self->{Inherits}->{$name} = $inh;

	my @errors;

	foreach ($inh->globals) {
		my $err = $self->global($_);
		push(@errors, $err), next if $err;
		# Variable flags? Accessibility.
	}

	foreach ($inh->methods) {
		my $err = $self->method($_);	# Check if efun
		push(@errors, $err), next if $err;
		$self->{MethodFlags}->{$_->name} =
				$inh->{MethodFlags}->{$_->name} | M_INHERITED;
	}

	return @errors;
}

# Debugging

sub dump {
	my ($self) = @_;

	my @inh = map { "(inherit " .
					quote(printable $_) . " " .
					quote(printable $self->{Inherits}->{$_}->path)
					. ")" }
					keys %{$self->{Inherits}};
	my @glob = map { $_->dump(1) } values %{$self->{Globals}};
	my @meth = map { $self->{Methods}->{$_}->dump(1) }
				grep { ! ($self->{MethodFlags}->{$_} & M_EFUN) }
					keys %{$self->{Methods}};

	my $out = "(program\n\t" . join("\n\t", @inh, @glob, @meth) . "\n)";

	return $out;
}

# Output

sub generate {
	my ($self, $path) = @_;

	my $package = $self->package;
	my $head = << "EOH";
# Program $path
package $package;
use strict;
EOH

	my @vars = qw($PROGRAM);
	my $base;

	if (scalar %{ $self->{Inherits} }) {
		my $inh = join " ",
				map { $_->package }
						values %{ $self->{Inherits} };
		push(@vars, qw(@ISA));
		$base = << "EOI";

\@ISA = qw($inh);
EOI

	}
	else {
		$base = << "EON";

sub new {
	my \$class = shift;
	my \$self = { };
	return bless \$self, \$class;
}
EON
	}

	$head .= 'use vars qw(' . join(" ", @vars) . ");\n";
	# XXX $path forms part of a Perl program. Beware.
	$head .= '*PROGRAM = \$' . __PACKAGE__ . "::PROGS{'$path'};\n";

	# These have a very large extent.
	local *::methods = $self->{Methods};
	local *::methodflags = $self->{MethodFlags};

	my @meth = map { $::methods{$_}->generate(0, $path) }
				grep { ! ($::methodflags{$_} & M_EFUN) }
					keys %::methods;

	my $tail = "\n1;\n";

	return $head . $base . join("\n\n", @meth) . $tail;
}

1;
